#include <cstdlib>
#include <cassert>


void* aligned_malloc(size_t size, int alignment)
{
    const int pointerSize = sizeof(void*);
    const int requestedSize = size + alignment - 1 + pointerSize;
    void* raw = std::malloc(requestedSize);
    void* start = (char*)raw + pointerSize;
    void* aligned = (void*)(((unsigned int)((char*)start+alignment-1)) & ~(alignment-1));  
    *(void**)((char*)aligned-pointerSize) = raw;
    return aligned;
}

void aligned_free(void* aligned)
{
    void* raw = *(void**)((char*)aligned-sizeof(void*));
    std::free(raw);
}


bool IsAligned(void* data, int alignment)
{
    assert((alignment & (alignment-1)) == 0);
    return ((uintptr_t)data & (alignment-1)) == 0;
}


#define DYNAMIC_ALIGNMENT(T) \
	void* operator new(size_t size) \
	{ \
		return aligned_malloc(size, __alignof(T)); \
	} \
	void* operator new(size_t size, int alignment) \
	{ \
		return aligned_malloc(size, alignment); \
	} \
	void* operator new[](size_t size) \
	{ \
		return aligned_malloc(size, __alignof(T)); \
	} \
	void* operator new[](size_t size, int alignment) \
	{ \
		return aligned_malloc(size, alignment); \
	} \
	void operator delete(void* p) \
	{ \
		aligned_free(p); \
	} \
	void operator delete[](void* p) \
	{ \
		aligned_free(p); \
	}


struct MyData
{
	int a;
	float b;
	int stuff[3];
	
	DYNAMIC_ALIGNMENT(MyData);
};


__declspec(align(64)) struct MyData2
{
	int a;
	float b;
	int stuff[3];
	
	DYNAMIC_ALIGNMENT(MyData2);
};



int main(int argc, char *argv[])
{
	assert(sizeof(MyData) == 20);
	
	MyData* data1 = new MyData;
	MyData* data2 = new (128) MyData;
	
	assert(IsAligned(data1, 4));
	assert(IsAligned(data2, 128));
	
	delete data2;
	delete data1;
	
	// By default, MyData2 will be aligned on 64-byte boundaries because of __declspect align
	MyData2* data3 = new MyData2;
	MyData2* data4 = new (256) MyData2;

	assert(IsAligned(data3, 64));
	assert(IsAligned(data4, 256));

	delete data4;
	delete data3;
	
	MyData* dataArray1 = new (256) MyData[5];
	MyData2* dataArray2 = new MyData2[10];
	
	assert(IsAligned(dataArray1, 256));
	assert(IsAligned(dataArray2, 64));
	
	delete[] dataArray1;
	delete[] dataArray2;
	
	return 0;
}

